Require Import Ensembles.
Require Import Exec.
Require Import relationOperators.
Require Import Partial_Order.
Require Import non_empty_list.

Module Type PROGRAM.
  Declare Module E:EXECUTIONS.
  Export E.
  Notation Exec := t.

  Open Scope type_scope.

  Module Type JUSTIFICATION.
    Parameter Execution : Type.
    Parameter Commitments : Type.
    Definition t := Execution * Commitments.

    Definition exec (x:t) := fst x.
    Definition comm (x:t) := snd x.
  End JUSTIFICATION.

  Module Justification <: JUSTIFICATION.
    Definition Execution := Exec.
    Definition Commitments := Ensemble Act.
    Definition t := prod Execution Commitments.
  
    Definition exec (x:t) := fst x.
    Definition comm (x:t) := snd x.
  End Justification.

  Module Commit_seq.
    Definition commitE :=
      sList Justification.t.
  End Commit_seq.
  
  Import Commit_seq.
  Import Justification.

  Parameter Program_Executions : Program -> Ensemble Exec.

  Implicit Arguments Ensembles.In [U].

  Notation eIn := Ensembles.In.
  
  (*
  (*
     * This axiom might be not needed.
     *)
  Axiom Program_Executions_match :
    forall (P:Program)(E:Exec), 
      eIn (Program_Executions P) E -> program E = P.
  *)


  (*
     * The Valid Executions of the JMM are those that are defined for this 
     * function, and to be so, it must satisfy all the axioms described below
     * for the Justifications function.
     *)
  (*
     * This function takes a program P, an execution E, and a proof that the 
     * execution is a valid execution  for that program, and it returns a 
     * justification for the execution.
     *)
  Parameter justification : 
    forall (P:Program)(E:Exec), 
      eIn (Program_Executions P) E -> commitE.

  Section justification_axioms.

    Implicit Arguments justification [P E].

    Variables (P:Program)(E:Exec).

    (*
       * All the actions in the execution must be justified.
       *)
    Definition allActionsAreJustified :=
      forall (E_exec_P:eIn (Program_Executions P) E)(j:Justification.t)(a:Act), 
        nth (justification E_exec_P) (limit (justification E_exec_P)) = Some j -> 
        actions E a ->
        eIn (comm j) a.

    (*
       * All the executions in a justification sequence are themselves executions 
       * of the program.
       *)
    Definition justificationsAreExecutions := 
      forall (E_exec_P:eIn (Program_Executions P) E)(j:Justification.t), 
	In (justification E_exec_P) j -> eIn (Program_Executions P) (exec j).

     (*
       * Commited actions of a CommitExec are actions of the execution being 
       * justified and the execution used to justify it.
       * <FIXME> This sounds like a tongue-twister. <\FIXME>
       *)
    Definition commitedActionInE :=
      forall (E_exec_P:eIn (Program_Executions P) E)(j:Justification.t)(a:Act), 
        In (justification E_exec_P) j -> 
        eIn (comm j) a ->
        eIn (actions E) a.

    Definition commitedActionInE' :=
      forall (E_exec_P:eIn (Program_Executions P) E)(j:Justification.t)(a:Act), 
        In (justification E_exec_P) j -> 
        eIn (comm j) a ->
        eIn (actions (exec j)) a. 

    (*
       * The commited actions in the Justification sequence have to be nested.
       * Pag 24 of the JMM POPL paper.
       * <NOTE>
       * In Pseudo-COQ:
       *   forall i:nat, i<lenght Justification(E) -> 
       *                    icluded (commits Justification(E)[i])
       *                            (commits Justification(E)[i+1]).
       * <\NOTE>
       *)
    
    Implicit Arguments Strict_Included [U].

    Definition commit_0_empty :=
      forall (E_exec_P:eIn (Program_Executions P) E)(j:Justification.t), 
        nth (justification E_exec_P) 0 = Some j -> 
        Justification.comm j = Ensembles.Empty_set _.

    Definition commitsAreNested :=
      forall (E_exec_P:eIn (Program_Executions P) E)(j j':Justification.t)(n:nat), 
        n < limit (justification (E_exec_P)) -> 
        nth (justification E_exec_P) n = Some j -> 
        nth (justification E_exec_P) (S n) = Some j' -> 
        Strict_Included (comm j) (comm j').

    (*
       * The justification sequence must commit all the actions in the executions 
       * being justified.
       * Pag 24 of the JMM POPL paper.
       *)
    Definition allActionsCommited :=
      forall (E_exec_P:eIn (Program_Executions P) E)(j:Justification.t), 
        nth (justification E_exec_P) (limit (justification E_exec_P))  = Some j -> 
        comm j = actions E.

    Section causality_requirements.
      Implicit Arguments Included [U].
      
      Variables x y:Act.

      (*
         * Forall C CommitExec of the Execution E, every 
         * action commited in C must be an action of the justifying execution 
         * in C.
         *)

      Definition req1 :=
        forall (E_exec_P: eIn (Program_Executions P) E)
          (j:Justification.t), 
          In (justification E_exec_P) j -> 
          Included (comm j) (actions (exec j)).

      (*
         * TODO: Show that this (below) is true.
         * I think this property should be replaced by the better property that 
         * po_i|C_i = po|C_i. In fact they are both equivalent when combined 
         * with the req1.
         *
         * The hb relation of the justifying execution is equal to the one being 
         * justified upto C_i.
         * <JMM> hb_i|C_i = hb|C_i <\JMM>
         *)

      Definition req2 := 
        forall (E_exec_P:eIn (Program_Executions P) E)(j:Justification.t)
          (x y:Act), 
          In (justification E_exec_P) j ->
          eIn (comm j) x-> 
          eIn (comm j) y -> 
          (hb E x y <-> hb (exec j) x y).

      (*
         * The so relation of the justifying execution is equal to the one being 
         * justified upto C_i.
         * <JMM> so_i|C_i = so|C_i <\JMM>
         *)
      
      Definition req3 :=
        forall (E_exec_P:eIn (Program_Executions P) E)(j:Justification.t)
          (x y:Act), 
          In (justification E_exec_P) j ->
          eIn (comm j) x-> 
          eIn (comm j) y -> 
          (so E x y <-> so (exec j) x y).

      (*
         * The V function of the justifying execution is equal to the one being 
         * justified upto C_i.
         * <JMM> V_i|C_i \in V|C_i <\JMM>
         *)
      Definition req4 :=
        forall (E_exec_P:eIn (Program_Executions P) E)(j:Justification.t)
          (w:Act)
          (w_act_E:actions E w)
          (w_act_j:actions (exec j) w)
          (w_write:act_kind w = Write), 
          In (justification E_exec_P) j ->
          eIn (comm j) w ->
          V E w w_act_E w_write = V (exec j) w w_act_j w_write.

      (*
         * The W function of the justifying execution is equal to the one being 
         * justified upto C_{i-1}.
         * <JMM> W_i|C_{i-1} = W|C_{i-1} <\JMM>
         *)
      
      Definition req5 :=
        forall (E_exec_P:eIn (Program_Executions P) E)(j j':Justification.t)(n:nat)
          (r:Act)
          (r_act_E:actions E r)
          (r_act_j:actions (exec j) r)
          (r_read:act_kind r = Read), 
          n < (limit (justification E_exec_P)) -> 
          nth (justification E_exec_P) (S n) = Some j ->
          nth (justification E_exec_P) n = Some j' ->
          eIn (comm j') r -> 
          W E r r_act_E r_read = W (exec j) r r_act_j r_read.
      
      (*
         * All reads not commited in a previous justifying execution see 
         * writes that happened before in this justifying execution.
         * forall r:Reads(A_i\C_{i-1}), hb_i W_i(r) r
         *)
      
      Definition req6 :=
        forall (E_exec_P:eIn (Program_Executions P) E)(j j':Justification.t)(n:nat)
          (r:Act)
          (r_act_j:actions (exec j) r)
          (r_read:act_kind r = Read), 
          n < (limit (justification E_exec_P)) -> 
          nth (justification E_exec_P) (S n) = Some j ->
          nth (justification E_exec_P) n = Some j' ->
          ~ eIn (comm j') r -> 
          hb (exec j) (W (exec j) r r_act_j r_read) r.

      Definition req6_comm0 :=
        forall (E_exec_P:eIn (Program_Executions P) E)(j:Justification.t)
          (r:Act)
          (r_act_j:actions (exec j) r)
          (r_read:act_kind r = Read),
          nth (justification E_exec_P) 0 = Some j ->
          hb (exec j) (W (exec j) r r_act_j r_read) r.
      
      Definition req7 :=
        forall (E_exec_P:eIn (Program_Executions P) E)(j j':Justification.t)(n:nat)
          (r:Act)
          (r_act_E:actions E r)
          (r_act_j:actions (exec j) r)
          (r_read:act_kind r = Read),
          n < (limit (justification E_exec_P)) -> 
          nth (justification E_exec_P) (S n) = Some j ->
          nth (justification E_exec_P) n = Some j' ->
          eIn (comm j) r -> ~ eIn (comm j') r -> 
          eIn (comm j') (W (exec j) r r_act_j r_read) /\ 
          eIn (comm j') (W E r r_act_E r_read).

      Definition req8 :=
        forall (E_exec_P:eIn (Program_Executions P) E)(j j':Justification.t)(n:nat)
          (x y z:Act),
          n < (limit (justification E_exec_P)) -> 
          nth (justification E_exec_P) (S n) = Some j ->
          nth (justification E_exec_P) n = Some j' ->
          eIn (comm j) z -> ~ eIn (comm j') z -> 
          ssw (exec j) x y -> 
          hb (exec j) y z -> 
          forall (m:nat)(j'':Justification.t),
            S n < m -> m <= (limit (justification E_exec_P)) -> 
            nth (justification E_exec_P) m = Some j'' ->
            sw (exec j'') x y.

    End causality_requirements.

    (* FIXME: Patch *)
    
    Definition Well_Formedness (P:Program)(E:Exec) := 
      forall (E_exec_P:eIn (Program_Executions P) E),
        no_WriteRead_inversion E (hb E) /\
        no_WriteRead_intromission E (hb E).
        

    Definition commitsWellFormed (P:Program)(E:Exec) := 
      forall (E_exec_P:eIn (Program_Executions P) E)(j:Justification.t), 
        In (justification E_exec_P) j -> 
        Well_Formedness P (exec j).
  
    (* END FIXME: Patch *)
  
  End justification_axioms.
  
  (*** FIXME : This should all be -> instead of /\, when doing that fix all the destruct 
     * clauses (may be with a tactic).
     *)
  Axiom ValidExecutions : forall (P:Program)(E:Exec), 
    allActionsCommited P E /\
    justificationsAreExecutions P E /\
    commitedActionInE P E /\
    commitedActionInE' P E /\
    commit_0_empty P E /\
    commitsAreNested P E /\
    req1 P E /\
    req2 P E /\ 
    req3 P E /\ 
    req4 P E /\ 
    req5 P E /\
    req6 P E /\
    req7 P E /\ 
    req6_comm0 P E /\
    commitsWellFormed P E /\
    Well_Formedness P E.

  Section program_properties.

    Section execution_properties.
      Variable E:Exec.

      (* 
       * the condition of ordering can be proved by the condition of
       * totality.
       *)
      Definition sequential_consistent : Prop :=
	exists to:Relation Act,
	  only_on_actions E to /\
	  Order to /\
	  total_upto to (actions E) /\
	  no_WriteRead_inversion E to /\
	  no_WriteRead_intromission E to /\
          rel_consistency (po E) to.
      
      Definition dataRace_free : Prop :=
	forall x y:Act, 
	  actions E x -> actions E y ->
	  act_var x = act_var y ->
	  act_kind x = Write \/ act_kind y = Write ->
	  hb E x y \/ hb E y x.

    End execution_properties.

    Definition correctly_synchronized (P:Program) : Prop :=
      forall E:Exec, 
	eIn (Program_Executions P) E -> 
	sequential_consistent E -> 
	dataRace_free E.

  End program_properties.

End PROGRAM.
