Require Import Relations.
Require Import Ensembles.
Require Import Partial_Order.
Require Import relationOperators.

Module Type DOMAIN.
  
  Parameter Program : Set.
  
  Inductive Kind : Set :=
  | Read
  | Write
  | VolatileRead
  | VolatileWrite
  | Lock
  | Unlock.
    
  Module Type GENERIC_TYPE.
    Parameter t : Set.
    
    Axiom t_dec : forall x y:t, {x = y} + {x <> y}.
  End GENERIC_TYPE.
  
  Declare Module Thread_ID : GENERIC_TYPE.
  Declare Module Uid : GENERIC_TYPE. 
  Declare Module VarName : GENERIC_TYPE. 
  Declare Module Value : GENERIC_TYPE.
  
  Module Type ACTIONS.
    Parameter t : Set. 
    
    Parameter act_kind : t -> Kind.
    Parameter act_thID : t -> Thread_ID.t.
    Parameter act_uid : t -> Uid.t.
    Parameter act_var : t -> VarName.t.
    Parameter synchronization : t -> bool.

    Axiom act_dec : forall x y:t, {x = y} + {x <> y}.
  End ACTIONS.
  Declare Module Actions : ACTIONS.

End DOMAIN.

Module Type EXECUTIONS.
  
  Declare Module Domain : DOMAIN. Import Domain.
  Export Domain.
  
  Notation Act := Actions.t.
  Notation Val := Value.t.
  Export Actions.
  
  Parameter t : Type.
  
  Parameter program : t -> Program.

  Parameter actions : t -> Ensemble Act.
  
  Implicit Arguments Order [U].

  Parameter po : t -> Relation Act.

  Parameter so : t -> Relation Act.

  Parameter W : 
    forall (E:t)(x:Act), 
      actions E x -> act_kind x = Read -> Act.

  Parameter V : 
    forall (E:t)(x:Act), actions E x -> act_kind x = Write -> Val.

  Definition sw (E:t) : Relation Act := 
    fun x y:Act => so E x y /\ act_var x = act_var y /\ 
      (act_kind x = Write /\ act_kind y = Read \/ 
      act_kind x = Unlock /\ act_kind y = Lock).

  Implicit Arguments clos_trans [A].

  Definition hb (E:t) : Relation Act :=
    clos_trans (fun x y:Act => sw E x y \/ po E x y).

  (*
    * The smallest relation that is included in sw and when composed with the
    * po gives hb as its transitive closure.
    *)
  Definition ssw (E:t)(x y:Act) :=
    forall (P: Relation Act), 
      (forall a b:Act, P a b -> sw E a b) ->
      (forall a b:Act, 
	hb E a b -> clos_trans (fun a b:Act => P a b \/ po E a b) a b) -> 
      P x y.

  Section auxiliar_definitions.

    Variable E:t.
    Variable rel:Relation Act.

    Definition only_on_actions : Prop :=
      forall x y:Act, rel x y -> actions E x /\ actions E y.

    (*
       * Given a relation over the Actions this definition states that 
       * Write-Seen function must be consistent with that order.
       * Used to specify so and po consistency.
       *)
    Definition no_WriteRead_inversion : Prop :=
      forall (r:Act)(act_r:actions E r)(read_r:act_kind r = Read), 
	 ~ rel r (W E r act_r read_r).

    (*
       * This definition states that for every read there can not be a write 
       * that comes after the write seen by that read and the read itself in 
	 * the order specified as argument.
       * Used to specify so and po consistency.
       *)
    Definition no_WriteRead_intromission : Prop :=
      forall (w r:Act)(act_r:actions E r)(read_r:act_kind r = Read), 
	act_var w = act_var r ->
	act_kind w = Write -> 
	(W E r act_r read_r) <> w -> 
	~ (rel (W E r act_r read_r) w /\ rel w r).
  
  End auxiliar_definitions.
  
  Section basic_axioms.
    
    Variable E:t.

    Axiom only_actions_in_E : 
      only_on_actions E (po E) /\ only_on_actions E (so E).

    Axiom po_partial_order : Order (po E).

    Implicit Arguments total_upto [A].

    (*
       * The po order is a total order of the actions of each single thread of 
       * the execution. Furthermore if two actions are related by po then both 
       * are in the same thread.
       *)
    Axiom po_total_upto_threadEq : 
      forall t: Thread_ID.t,
	total_upto (po E) (fun x:Act => act_thID x = t).
    
    Axiom po_then_threadEq :
      forall x y:Act, po E x y \/ po E y x -> act_thID x = act_thID y.

    Axiom so_partial_order : forall E:t, Order (so E).
  
      (*
       * The so order is a total order of the actions of the execution. 
       * Furthermore if two actions are related by so then both are 
       * synchronization actions.
       *)
    Axiom so_total_upto_synchronization : 
      total_upto (so E) (fun x:Act => synchronization x = true).
   
    Axiom so_then_synchronization :
      forall x y:Act, 
	so E x y \/ so E y x -> 
	synchronization y = true /\ synchronization x = true.

    Axiom so_po_consistency : rel_consistency (so E) (po E).

    (*
       * so is consistent with the write-seen funciton.
       * <FIXME>
       * This should only relate Volatile Variables. Though the definition 
       * remains the same. (I could simply make a single definition for 
       * Volatiles and non-Volatiles)
       * <\FIXME>
       *)
    Axiom synch_ord_consistent : 
	no_WriteRead_inversion E (so E) /\ 
	no_WriteRead_intromission E (so E).
	
    (*
       * hb is consistent with the write-seen funciton.
       *)
    Axiom hb_ord_consistent : 
      no_WriteRead_inversion E (hb E) /\
      no_WriteRead_intromission E (hb E).

    Axiom w_returns_writes : 
      forall (r:Act)(act_r:actions E r)(r_read:act_kind r = Read), 
        act_kind (W E r act_r r_read) = Write.

    Axiom w_return_actions :
      forall (r:Act)(act_r:actions E r)(r_read:act_kind r = Read), 
        actions E (W E r act_r r_read).

    Axiom w_on_same_var : 
      forall (r:Act)(act_r:actions E r)(r_read:act_kind r = Read), 
        act_var (W E r act_r r_read) = act_var r.
  

  End basic_axioms.

  Implicit Arguments In [U].

  Section comparing_executions.
    
    Variables E E':t.
    
    Implicit Arguments In [U].
    
    Variable set : Ensemble Act.

    Definition eq_act_upto :=
      forall x:Act, In set x -> actions E x <-> actions E' x.

    Hypothesis eq_act_upto1 :
      forall x:Act, In set x -> actions E x <-> actions E' x.

    Definition eq_ord_act (rel:t -> Relation Act)(x:Act) :=
      forall y:Act, rel E y x <-> rel E' y x.

    Definition eq_ord_upto (rel: t -> Relation Act) : Prop := 
      forall x y, In set x -> In set y -> 
        (rel E x y <-> rel E' x y).

    Definition act_both (x:Act)(_ : In set x)(h:actions E x) : actions E' x :=
      let i := eq_act_upto1 x in
        match i with
          | conj E_E' _ => E_E' (fun _ : In set x => h)
        end.

    Definition eq_W :=
      forall (r:Act)(r_act_E:actions E r)(r_act_E':actions E' r)
        (r_read:act_kind r = Read), In set r -> 
          W E r r_act_E r_read = W E' r r_act_E' r_read.
    

    Definition eq_W2 := 
      forall (r:Act)(r_action:actions E r)(r_read:act_kind r = Read)
        (h:In set r), 
        W E r r_action r_read = W E' r (act_both r h r_action) r_read.
    
    Definition eq_V :=
      forall (w:Act)(w_act_E:actions E w)(w_act_E':actions E' w)
        (w_write: act_kind w = Write),
        In set w -> 
          V E w w_act_E w_write = V E' w w_act_E' w_write.
  
  End comparing_executions.
  
  Definition eq_Exec_ord_upto (E E':t)(set:Ensemble Act) :=
    eq_ord_upto E E' set po /\
      eq_ord_upto E E' set so.
  
  Definition eq_Exec_func_upto (E E':t)(set:Ensemble Act) :=
    eq_W E E' set /\
      eq_V E E' set.
  
  (*
     * For every element in the set all the lower elements than that in a certain
     * order are also in the set.
     *)
  Definition complete_upto (set:Ensemble Act)(rel:Relation Act) :=
    forall x y:Act, In set y ->
      rel x y -> In set x.

  Definition same_actions (E E':t)(set:Ensemble Act) :=
    forall x:Act, 
      In set x -> actions E x <-> actions E' x.

  Definition eq_Exec_complete_upto (set_ord set_func:Ensemble Act)(E E':t) :=
    eq_act_upto E E' set_ord /\
    eq_Exec_ord_upto E E' set_ord /\ 
    eq_Exec_func_upto E E' set_func /\ 
    complete_upto set_ord (po E) /\ 
    complete_upto set_ord (po E') /\
    complete_upto set_ord (so E) /\ 
    complete_upto set_ord (so E').

End EXECUTIONS.

Module ExecutionFacts (E:EXECUTIONS).
  Export E.

  Section lemmas.

    Variable E:t.

    Lemma po_single_thread : 
      forall x y:Act, 
	act_thID x = act_thID y <-> (po E x y \/ po E y x).
    Proof.
      intros x y; split.
      
      intro same_thread_x_y;
	apply ((po_total_upto_threadEq E (act_thID x)) x y);
	  [auto|exact (sym_eq same_thread_x_y)].
      
      exact (po_then_threadEq E x y).
    Qed.
  
    Lemma po_same_thread : 
      forall x y:Act, 
	po E x y -> 
	act_thID x = act_thID y.
    Proof.
      intros x y po_x_y.
      destruct (po_single_thread x y) as (_ , po_same_thread); 
	apply(po_same_thread (or_introl (po E y x) po_x_y)).
    Qed.

    Lemma same_thread_po : 
      forall x y:Act,
	act_thID x = act_thID y -> 
	po E x y \/ po E y x.
    Proof.
      intros x y x_thEq_y.
      destruct (po_single_thread x y) as (same_th_po, _); 
	apply (same_th_po x_thEq_y).
    Qed.

    Lemma so_only_on_synch : 
      forall x y:Act, so E x y \/ so E y x <-> 
	(synchronization x = true /\ synchronization y = true).
    Proof.
      intros x y; split.
      
      intro so_related_x_y.
      destruct (so_then_synchronization E x y so_related_x_y) 
	as (synch_y, synch_x); auto.

      intros synch_x_y; destruct synch_x_y as (synch_x, synch_y).
      apply ((so_total_upto_synchronization E) x y); auto.
    Qed.

    Lemma so_synchronization : 
      forall x y:Act, so E x y -> 
	(synchronization x = true /\ synchronization y = true).
    Proof.
      intros x y so_x_y.
      destruct (so_only_on_synch x y) as (so_synch, _); 
	apply (so_synch (or_introl (so E y x) so_x_y)).
    Qed.

    Lemma synch_so : 
      forall x y:Act,
	(synchronization x = true /\ synchronization y = true) ->
	so E x y \/ so E y x.
    Proof.
      intros x y synch_x_y.
      destruct (so_only_on_synch x y) as (_ , synch_so); 
	apply (synch_so synch_x_y).
    Qed.

    Lemma hb_only_on_actions:
      forall x y:Act, hb E x y -> 
        actions E x /\ actions E y.
    Proof.
      intros x y hb_x_y.
      induction hb_x_y as [x y base_case| 
        x y z hb_x_y ind_case_x_y hb_y_z ind_case_y_z].

      destruct (only_actions_in_E E) as (po_only_act, so_only_act).
      elim base_case; [intro sw_x_y| 
        intro po_x_y; exact (po_only_act x y po_x_y)].
      unfold sw in sw_x_y; elim sw_x_y; intros so_x_y _;
        exact (so_only_act x y so_x_y).

      intuition.
    Qed.

    Section eq_props.
      Variable E':t.
      Variable set: Ensemble Act. 
      
      Hypothesis eq_po_EE' : eq_ord_upto E E' set po.
      Hypothesis eq_so_EE' : eq_ord_upto E E' set so.

      Lemma eq_sw_EE' : eq_ord_upto E E' set sw.
      Proof.
        split.
        intro sw_x_y; constructor.
        unfold sw in sw_x_y; generalize (eq_so_EE' x y H H0); intuition. 
        unfold sw in sw_x_y; generalize (eq_so_EE' x y H H0); intuition. 
       
        intro sw'_x_y; constructor.
        unfold sw in sw'_x_y; generalize (eq_so_EE' x y H H0); intuition. 
        unfold sw in sw'_x_y; generalize (eq_so_EE' x y H H0); intuition.
      Qed.
      
      Lemma compl_trans_comp : 
        forall rel:relation Act, 
          complete_upto set rel -> 
            complete_upto set (trans_clos Act rel).
      Proof.
        intros rel comp_rel.
        unfold complete_upto; intros x y in_y trans_x_y.
        induction trans_x_y as [x y base| x z y ind];
          [exact (comp_rel x y in_y base)|
            exact (comp_rel x z (IHtrans_x_y in_y) ind)].
      Qed.

      Lemma compl_trans_comp1 : 
        forall rel:relation Act, 
          complete_upto set rel -> 
            complete_upto set (clos_trans rel).
      Proof.
        intros rel comp_rel.
        unfold complete_upto; intros x y in_y trans_x_y.
        induction trans_x_y as [x y base| x z y _ hip_ind_z_x _ hip_ind_y_z];
          [exact (comp_rel x y in_y base)|auto].
      Qed.

      Lemma eq_trans : 
        forall rel : t -> relation Act, 
          eq_ord_upto E E' set rel ->
            complete_upto set (rel E) -> 
              complete_upto set (rel E') -> 
                eq_ord_upto E E' set (fun e:t => (trans_clos Act (rel e))).
      Proof.
        intros rel eq_rel comp_rel comp_rel'.
        unfold eq_ord_upto.
        intros x y in_x in_y.
        split.

        intros tc_x_y.
        induction tc_x_y as [x y base | x z y rel_x_z tc_z_y hip_ind].
        constructor 1; assert (h:=eq_rel x y in_x in_y); intuition.

        assert (in_z:=(compl_trans_comp (rel E) comp_rel) z y in_y tc_z_y).
        constructor 2 with (y:=z);
          assert (h:=eq_rel x z in_x in_z); intuition.

        intros tc'_x_y.
        induction tc'_x_y as [x y base | x z y rel'_x_z tc'_z_y hip_ind].
        constructor 1; assert (h:=eq_rel x y in_x in_y); intuition.

        assert (in_z:=(compl_trans_comp (rel E') comp_rel') z y in_y tc'_z_y).
        constructor 2 with (y:=z);
          assert (h:=eq_rel x z in_x in_z); intuition.
      Qed.

      Lemma eq_trans1 : 
        forall rel : t -> relation Act, 
          eq_ord_upto E E' set rel ->
            complete_upto set (rel E) -> 
              complete_upto set (rel E') -> 
                eq_ord_upto E E' set (fun e:t => (clos_trans (rel e))).
      Proof.
        intros rel eq_rel comp_rel comp_rel'.
        unfold eq_ord_upto.
        intros x y in_x in_y.
        split.

        intros tc_x_y.
        induction tc_x_y as [x y base | x z y rel_x_z hip_ind_x_z tc_z_y hip_ind_z_y].
        constructor 1; assert (h:=eq_rel x y in_x in_y); intuition.

        assert (in_z:=(compl_trans_comp1 (rel E) comp_rel) z y in_y tc_z_y).
        constructor 2 with (y:=z);
          assert (h:=eq_rel x z in_x in_z); intuition.

        intros tc'_x_y.
        induction tc'_x_y as [x y base | x z y rel'_x_z hip_ind_x_z tc'_z_y hip_ind_z_y].
        constructor 1; assert (h:=eq_rel x y in_x in_y); intuition.

        assert (in_z:=(compl_trans_comp1 (rel E') comp_rel') z y in_y tc'_z_y).
        constructor 2 with (y:=z);
          assert (h:=eq_rel x z in_x in_z); intuition.
      Qed.


      Lemma eq_union_po_sw : 
        forall (rel rel':t -> relation Act),
          eq_ord_upto E E' set rel ->
            eq_ord_upto E E' set rel' ->
              eq_ord_upto E E' set (fun e:t => (fun x y:Act => rel e x y \/ rel' e x y)).
      Proof.
        intros rel rel' eq_rel eq_rel'.
        unfold eq_ord_upto.
        intros x y in_x in_y.

        split.

        intros union_x_y.
        destruct union_x_y as [rel_x_y| rel'_x_y].
        left; assert (h:=(eq_rel x y in_x in_y)); 
          destruct h as (h, _); auto.
        right; assert (h:=(eq_rel' x y in_x in_y)); 
          destruct h as (h, _); auto.
        
        intros union'_x_y.
        destruct union'_x_y as [rel_x_y| rel'_x_y].
        left; assert (h:=(eq_rel x y in_x in_y));
          destruct h as (_, h); auto.
        right; assert (h:=(eq_rel' x y in_x in_y)); 
          destruct h as (_, h); auto.
      Qed.

      Lemma comp_so_sw : 
        forall E:t,
          complete_upto set (so E) -> complete_upto set (sw E).
      Proof.
        intros E0 com_so.
        unfold complete_upto; intros x y in_y sw_x_y.
        destruct sw_x_y as (so_x_y, _).
        exact (com_so x y in_y so_x_y).
      Qed.
        
      Lemma comp_union :
        forall (E:t)(rel rel':t -> relation Act), 
          complete_upto set (rel E) -> 
          complete_upto set (rel' E) -> 
            complete_upto set 
              (fun x y:Act => rel E x y \/ rel' E x y). 
      Proof.
        intros E0 rel rel' comp_rel comp_rel'.
        unfold complete_upto; intros x y in_y rel_x_y.
        elim rel_x_y; clear rel_x_y; intros rel_x_y;
          [exact (comp_rel x y in_y rel_x_y) |
            exact (comp_rel' x y in_y rel_x_y)].
      Qed.
            
      Lemma eq_hb_EE' : 
        complete_upto set (po E) -> 
          complete_upto set (po E') ->   
            complete_upto set (so E) -> 
              complete_upto set (so E') ->
                eq_ord_upto E E' set hb.
      Proof.
        intros comp_po_R comp_po_E' comp_so_E comp_so_E'.
      
        unfold hb.
        apply (eq_trans1 (fun E0:t => (fun x y : Act => sw E0 x y \/ po E0 x y))).

        apply (eq_union_po_sw sw po).
        exact eq_sw_EE'. 
        exact eq_po_EE'.
        apply(comp_union E sw po); [exact(comp_so_sw E comp_so_E)|auto].
      
        apply (comp_union E' sw po); [exact(comp_so_sw E' comp_so_E')|auto].
      Qed.

      Lemma comp_hb : 
        complete_upto set (po E) -> 
          complete_upto set (so E) -> 
            complete_upto set (hb E).
      Proof.
        intros comp_po comp_so.
        unfold complete_upto.
        intros x y in_y hb_x_y.
        induction hb_x_y; [|auto].
        elim H; 
          [exact (comp_so_sw E comp_so x y in_y)|
          exact (comp_po x y in_y)].
      Qed.

    End eq_props.
  End lemmas.
End ExecutionFacts.
