Require Import List.
Require Import Classical.
Require Import Omega.

Section sList.
  Variable A B:Type.
  
  Set Implicit Arguments.
  
  Inductive sList (A:Type) : Type := 
  | sCons : A -> sList A
  | cons : A -> sList A -> sList A.
  
  Fixpoint length (l:sList A) : nat :=
    match l with 
      | sCons _ => 1
      | cons _ ls => S (length ls)
    end.
  
  Definition limit (l:sList A) : nat :=
    (length l) - 1.
  
  Infix "::" := cons (at level 60, right associativity) : sList_scope.
  
  Fixpoint app (l m:sList A) {struct l} : sList A :=
    match l with
      | sCons x => cons x m
      | cons x xs => cons x (app xs m)
    end.
  
  Infix "++" := app (at level 60, right associativity) : sList_scope.
  
  Open Scope sList_scope.
  
  Definition head (l:sList A) : A :=
    match l with 
      | sCons x => x 
      | x :: _ => x 
    end.
  
  Definition tail (l:sList A) : option (sList A) :=
    match l with 
      | sCons _ => None
      | _ :: ls => Some ls
    end.
  
  Fixpoint In (l:sList A)(a:A) {struct l} : Prop := 
    match l with 
      | sCons x => a = x
      | x :: xs => a = x \/ In xs a
    end.

  Fixpoint nth (l:sList A)(n:nat) {struct l} : option A :=
    match l, n  with 
      | sCons x, 0 => Some x
      | sCons x, S m => None
      | x :: _, 0 => Some x
      | _ :: ls, S m => nth ls m
    end.

  Fixpoint rev (l:sList A) : sList A :=
    match l with 
      | sCons x => sCons x
      | x :: xs => (rev xs) ++ sCons x
    end.

  Fixpoint map (l:sList A)(func:A -> B) {struct l} : sList B :=
    match l with 
      | sCons x => sCons (func x)
      | x::xs => (func x)::(map xs func)
    end.

  Inductive mapProp (p:A -> Prop) : sList A -> Prop :=
    | mp_cons : forall x:A, p x -> mapProp p (sCons x)
    | mp_ind : forall (x:A)(l:sList A), p x /\ mapProp p l -> mapProp p (x::l).

  Fixpoint mapBinProp (l:sList A)(func:A -> A -> Prop) {struct l} : Prop :=
    match l with 
      | sCons _ => True
      | x::xs => 
        match xs with 
          | sCons y => func x y
          | y::ys => func x y /\ mapBinProp xs func
        end
    end.

  Lemma length_positive : forall l:sList A,  1 <= length l.
  Proof.
    induction l; [auto| simpl; omega].
  Qed.

  Lemma S_limit : forall (l:sList A)(a:A), limit (a::l) = S (limit l).
  Proof.
    unfold limit; intros l a; simpl;
      rewrite (Minus.minus_Sn_m _ _ (length_positive l)); reflexivity.
  Qed.

  Lemma nth_Some : 
    forall (ls:sList A)(i:nat),
      i <= limit ls -> 
      exists a:A, nth ls i = Some a.
  Proof.
    induction ls.

    intros i i_leq_lim;
      compute in i_leq_lim;
        induction i; [exists a; auto| 
          assert(h:=Le.le_n_O_eq _ i_leq_lim); discriminate h].

    intros i i_leq_lim;
      induction i; [exists a; compute; reflexivity|].
    simpl;
      apply IHls; 
        rewrite (S_limit ls a) in i_leq_lim; omega.
  Qed.

  Lemma some_Nth : 
    forall (ls:sList A)(i:nat)(a:A),
      nth ls i = Some a ->
      i <= limit ls.
  Proof.
    induction ls; destruct i; 
      [intros; omega
        |intros a0 nth_cons_a0; discriminate nth_cons_a0|
          intros; omega|].
    intros a0 nth_cons_a0; simpl in nth_cons_a0 |- *; 
      rewrite (S_limit ls a).
    exact (le_n_S _ _ (IHls i a0 nth_cons_a0)).
  Qed.


  Lemma nth_In : forall (n:nat)(ls:sList A)(a:A), 
    nth ls n = Some a -> 
    In ls a.
  Proof.
    induction n; destruct ls;
      [simpl; intros; injection H; auto
        | simpl; intros; injection H; auto 
        | simpl; intros; discriminate H
        |simpl; intros; right; exact(IHn _ _ H)].
  Qed.

  Lemma In_nth : forall (ls:sList A)(a:A), 
    In ls a -> exists n:nat, nth ls n = Some a.
  Proof.
    intros ls a in_ls_a.
    
    induction ls.
    compute in in_ls_a; exists 0; rewrite in_ls_a; auto.
    
    destruct in_ls_a as [a_a0|in_ls_a]; 
      [exists 0; simpl; rewrite a_a0; auto
        |destruct (IHls in_ls_a) as (n , nth_ls_n_a); 
          exists (S n); compute; auto].
  Qed.

  Lemma nth_limit_In : 
    forall (ls:sList A)(a:A), 
      nth ls (limit ls) = Some a -> 
      In ls a.
  Proof fun (ls:sList A) => nth_In (limit ls) ls.
  
  Lemma singleFail_mapFail : forall (l:sList A)(a:A)(P:A -> Prop), 
    In l a -> 
    ~ P a -> 
    ~ mapProp P l.
  Proof.
    intros l a P a_in_l not_pa; induction l.

    red; intro mp_l; inversion mp_l; simpl in a_in_l; subst; contradiction.

    red; intro mp_l.
    elim (classic (a = a0));
      [ intro a_eq_a0; subst; inversion mp_l;
        decompose [and] H0; contradiction
        | elim a_in_l; [auto|
          intros in_l_a a_not_a0; inversion mp_l; decompose [and] H0;
            exact (IHl in_l_a H3)
        ]
      ].
  Qed.
    
  Lemma failInd : forall (l:sList A)(a:A)(P:A -> Prop), 
    P a -> ~ mapProp P (a::l) ->  ~ mapProp P l.
  Proof.
    intros l a P P_a n_mp_al.
    elim (classic (mapProp P l)); 
      [intro mp_l; 
        assert (mp_al:= mp_ind _ (conj P_a mp_l)); 
          contradiction
        |auto].
  Qed.

  Lemma firstFail : 
    forall (P:A -> Prop)(l:sList A), ~ mapProp P l-> 
      exists i:nat, i <= limit l /\
        (forall a:A, nth l i = Some a -> ~ P a) /\ 
        (forall (j:nat)(a:A), 
          j < i -> nth l j = Some a -> P a).
  Proof with auto.
    intros P l not_mp_l.
    induction l.

    elim (classic (P a));
      [intro P_a; assert (mp_a:=mp_cons _ _ P_a); contradiction|
        intro n_Pa; exists 0;
          split; [auto |
            split; [intros a0 sa_eq_sa0; assert(a_eq_a0:a=a0); 
              [injection sa_eq_sa0;auto|subst;auto ] |
                intros j a0 s_j_leq_0; 
                  assert(h:= Le.le_n_O_eq (S j) s_j_leq_0); discriminate h
            ]
          ]
      ].
    
    elim (classic (P a)).

    intro P_a.
    destruct (IHl (failInd P_a not_mp_l)) as (x, H); clear IHl;
      destruct H as (x_leq_lim, H);
        destruct H as (n_Pa_x, lower_P).

    exists (S x);
      split; [rewrite (S_limit l a);exact (Le.le_n_S _ _ x_leq_lim) |
        split; [simpl; auto|
          intro j; case j; [compute;intros a0 _ sa_eq_sa0;
            assert (a = a0); injection sa_eq_sa0; subst; auto|
              intros n a0 Sn_leq_Sx nth_l_sx_a0; 
                apply (lower_P n a0); [exact (Lt.lt_S_n _ _ Sn_leq_Sx)|
                  simpl in nth_l_sx_a0; auto]
          ]
        ]
      ].

    intro n_Pa; exists 0.

    split; [omega|
      split; [compute; intros a0 sa_eq_sa0; assert (a = a0); 
        [injection sa_eq_sa0| subst]; auto |
          intros j a0 j_lt_0 _; assert (h:=Le.le_n_O_eq _ j_lt_0); 
            discriminate h
      ]
    ].
  Qed.

End sList.
