Require Import Comm.
Require Import relationOperators.
Require Import jmmProperties.
Require Import Partial_Order.
Require Import Classical.

Module ProgramProofs (P:PROGRAM).

  Module PF := ProgramFacts(P).
  Import PF.
  Import P.
  
  Section proof.
    
    Variable E:Exec.
    Hypothesis reads_see_hb :
      forall (r:Act)(r_in_E:actions E r)(r_read:act_kind r = Read), 
        hb E (W E r r_in_E r_read) r.
    
    Definition to := top_sort (hb E) (hb_partial_order E).
    
    Lemma hb_to : 
      forall x y:Act, x <> y -> hb E x y -> to x y.
    Proof.
      intros x y x_not_y hb_x_y. 

      assert (in_hb_x: in_rel (hb E) x).
      exists y; auto.

      assert (in_hb_y: in_rel (hb E) y).
      exists x; auto.

      destruct (top_sort_total (hb E)(hb_partial_order E) x y in_hb_x in_hb_y) as 
        [to_x_y| to_y_x]; auto.
      assert (h:=(top_sort_consistent _ (hb_partial_order _)) _ _ hb_x_y to_y_x); 
        contradiction.
    Qed.

   Lemma to_hb_only_on_actions : 
      only_on_actions E to.
    Proof.
      red; intros x y to_x_y.
      destruct (ts_only_on_r (hb E)(hb_partial_order E) x y to_x_y) as (in_hb_x, in_hb_y).
      
      split.
      
      destruct in_hb_x as (x0, hb_rel_x0_x); destruct hb_rel_x0_x as [hb_x_x0|hb_x0_x];
        [exact (proj1 (hb_only_on_actions E x x0 hb_x_x0))|
           exact (proj2 (hb_only_on_actions E x0 x hb_x0_x))].

      destruct in_hb_y as (x0, hb_rel_x0_y); destruct hb_rel_x0_y as [hb_y_x0|hb_x0_y];
        [exact (proj1 (hb_only_on_actions E y x0 hb_y_x0))|
           exact (proj2 (hb_only_on_actions E x0 y hb_x0_y))].
    Qed.

    Lemma to_order :
      Order to.
    Proof (top_sort_order (hb E) (hb_partial_order E)).

    Lemma to_total : total_upto to (actions E).
    Proof.
      red; intros x y x_act y_act.

      assert (in_rel (hb E) x) as in_rel_x.
      exact (ex_intro _ _ (or_introl _ (hb_reflexive _ x))).

      assert (in_rel (hb E) y) as in_rel_y.
      exact (ex_intro _ _ (or_introl _ (hb_reflexive _ y))).
      
      exact (top_sort_total _ _ _ _ in_rel_x in_rel_y).
    Qed.

    Lemma to_no_wr_inversion :
      no_WriteRead_inversion E to.
    Proof.
      intros r act_r read_r.
      assert (hb_w_r:= reads_see_hb _ act_r read_r).

      assert (write_w:= w_returns_writes _ _ act_r read_r).
      red; intros to_w_r;
        rewrite ((top_sort_consistent (hb _) (hb_partial_order _) _ _ hb_w_r) to_w_r) in write_w;
          rewrite read_r in write_w; discriminate.
    Qed.

    Lemma to_po_consistent : rel_consistency (po E) to.
    Proof.
      intros x y po_x_y to_y_x.
      elim (classic (x = y)); auto; intro x_not_y.
      assert (to_x_y:= 
        hb_to _ _ x_not_y        
        (Relation_Operators.t_step _ (fun x y:_ => (sw _) x y \/ (po _) x y) _ _ (or_intror (sw E _ _) po_x_y))).
      destruct (top_sort_order _ (hb_partial_order E)) as (_, _, antisym);
        exact (antisym _ _ to_x_y to_y_x).
    Qed.

    Lemma to_wr_intromission : 
      no_WriteRead_intromission E to -> sequential_consistent E.
    Proof.
      intros.
      unfold sequential_consistent.
      exists to.
      split; [exact to_hb_only_on_actions|
        split; [exact (to_order)|
          split; [exact (to_total) | 
            split; [exact to_no_wr_inversion|
              split; [auto| exact to_po_consistent]
            ]
          ]
        ]
      ].
    Qed.

    Section first_error.

      Definition write_read_intromission_w (rel:Relation Act)(r:Act)
        (act_r:actions E r)(r_read:act_kind r = Read)(w:Act) : Prop :=
        actions E w /\
        act_kind w = Write /\
        act_var r = act_var w /\
        W E r act_r r_read <> w /\
        rel (W E r act_r r_read) w /\
        rel w r.
      
      Definition write_read_intromission (rel:Relation Act)(r:Act)
        (act_r:actions E r)(r_read:act_kind r = Read) : Prop :=
        exists w:Act, 
          write_read_intromission_w rel r act_r r_read w.

      Definition first_intromission (rel:Relation Act)(r:Act)
        (act_r:actions E r)(r_read:act_kind r = Read) : Prop :=
        write_read_intromission rel r act_r r_read /\
        forall (r':Act)(act_r':actions E r')(r'_read:act_kind r' = Read), 
          write_read_intromission rel r' act_r' r'_read -> 
          ~ rel r' r.


      (*
         * The conditions before the forall quantifier guarantee that there
         * are variables proofs that match the quatified variables.
         *)
      Axiom there_is_a_first_intromission : 
        ~ no_WriteRead_intromission E to -> 
        exists r:Act, 
          actions E r /\
          act_kind r = Read /\
          forall (act_r:actions E r)(r_read:act_kind r = Read), 
            first_intromission to r act_r r_read.

      (* 
         * TODO: It can be easily proved the uniqueness, since to is total. 
         *)
    
    End first_error.
    
    Section not_seq_facts.
      Hypothesis e_not_seq : ~ sequential_consistent E.
      Variable P:Program.
      Hypothesis p_well_synch : correctly_synchronized P.
      
      Lemma intromission : ~ no_WriteRead_intromission E to.
      Proof.
        red; intro introm;
          assert (e_seq:=to_wr_intromission introm); contradiction.
      Qed.

      Section axiom_sequential_continuation.
        Variable r w:Act.
        Variable act_r: actions E r.
        Variable r_read: act_kind r = Read.

        Inductive lower_r : Ensemble Act:=
          lower : forall x:Act, x <> r -> to x r -> In lower_r x.

        Inductive lower_eq_r : Ensemble Act:=
          lower_eq : forall x:Act, to x r -> In lower_eq_r x.

        (*
           Implicit Arguments Union [U].
           Implicit Arguments Singleton [U].
           *)

        Axiom sequential_exec : 
          Program_Executions P E ->
          first_intromission to r act_r r_read ->
          write_read_intromission_w to r act_r r_read w -> 
          exists E':Exec, 
            Program_Executions P E' /\
            eq_Exec_complete_upto lower_eq_r lower_r E E' /\
            (forall act_r_E':actions E' r, W E' r act_r_E' r_read = w) /\
            sequential_consistent E'.
      
      End axiom_sequential_continuation.

      Section race_evidence.
        
        Variables w r:Act.
        Variables (r_act:actions E r)(r_read:act_kind r = Read).
        Definition W : Act := W E r r_act r_read.
        Hypothesis w_write : act_kind w = Write.
        Hypothesis w_r_same_var : act_var w = act_var r.
        Hypothesis w_not_Wr : w <> W.
        Hypothesis introm_w : 
          to W w /\ to w r.
        
        Lemma at_least_data_race : ~ (hb E W w /\ hb E w r).
        Proof.
          destruct (hb_ord_consistent E) as (_, no_intro_hb).
          exact (no_intro_hb w r r_act r_read w_r_same_var w_write 
            (sym_not_equal w_not_Wr)).
        Qed.

        Lemma W_act : actions E W.
        Proof.
          elim (hb_only_on_actions E W r (reads_see_hb r r_act r_read)); auto.
        Qed.

        Lemma w_act : actions E w.
        Proof.
          assert (to_w_r:to w r); elim introm_w; auto; clear introm_w.

          destruct ((ts_only_on_r (hb E) (hb_partial_order E) w r) to_w_r) as 
            (in_rel_w, _).
          elim in_rel_w; clear in_rel_w; intros x hb_related_x_w _ _.

          elim hb_related_x_w; 
            [intro hb_w_x;elim (hb_only_on_actions E w x hb_w_x);auto|
              intro hb_x_w;elim (hb_only_on_actions E x w hb_x_w);auto].
        Qed.
        
        Lemma W_w_same_var : act_var W = act_var w.
        Proof.
          unfold W; rewrite (w_on_same_var E r r_act r_read); auto.
        Qed.

        Lemma or_elim : forall P Q:Prop, P \/ Q -> ~ P -> Q.
        Proof.
          intros p q p_or_q not_p.
          elim p_or_q; [intro; contradiction|auto].
        Qed.
          
        Lemma not_dataRace_free : ~ dataRace_free E.
        Proof.
          red; intros DRF.

          elim (not_and_or _ _ at_least_data_race).

          assert (w_W_writes:= or_intror (act_kind W = Write) w_write). 

          assert(hb_W_w:=DRF _ _ W_act w_act W_w_same_var w_W_writes);
            intro not_hb_W_w; assert(hb_w_W:=or_elim _ _ hb_W_w not_hb_W_w);
              clear hb_W_w not_hb_W_w.
          
          assert(w_eq_W:=top_sort_consistent (hb _) (hb_partial_order _) 
            _ _ hb_w_W (proj1 introm_w)); contradiction.

          elim (DRF _ _ w_act r_act w_r_same_var (or_introl (act_kind r = Write) w_write));
            auto.

          intro hb_r_w.
          rewrite <- (top_sort_consistent (hb _) (hb_partial_order _) 
            _ _ hb_r_w (proj2 introm_w)) in w_write.
          rewrite r_read in w_write; discriminate.
        Qed.

      End race_evidence.

    (*
       * Up to here we know there must be a read write intromission, we must
       * look for it and reverse the W function at that poin. After that, we
       * have to show that there is a sequential consistent continuation for 
       * that execution with that prefix, and later we arrive at a 
       * contradiction, since the writes are not hb ordered and thus, there is 
       * a data race in a seq. cons. execution of a correctly synch. program.
       *)

    End not_seq_facts.

    Section Lemma2_hypothesis.

      Variable P:Program.
      Hypothesis E_exec_P: Program_Executions P E.
      Hypothesis p_well_synch : correctly_synchronized P.

      Theorem Lemma2 : sequential_consistent E.
      Proof.

        (*By absurdity we assume E is not correctly synchrnonized*)
        elim (classic (sequential_consistent E)); [auto|]; intro not_seq_cons_E.

        destruct (there_is_a_first_intromission (intromission not_seq_cons_E)) as
          (r, (act_r, (r_read, first_intro)));
              assert(first_intro_r:= first_intro act_r r_read); 
                clear first_intro.
        
        elim first_intro_r; intros w_r_intro _;
          destruct w_r_intro as (w, intro_r_w).

        (* Looking for a data race in E *)
        elim intro_r_w; intros act_w w_intro_to.
        
        destruct w_intro_to as (w_write, (same_var_r_w, (Wr_not_w, 
          (to_Wr_w, to_w_r)))).

        assert(race_E:=(at_least_data_race _ _ act_r r_read w_write  
          (sym_eq same_var_r_w) (sym_not_eq Wr_not_w))).
        
        
        (* With this we get the E' execution *)
        elim ((sequential_exec P r w act_r r_read) E_exec_P first_intro_r intro_r_w);
          intros E' E'_seq_exec;
            destruct E'_seq_exec as (E'_exec_P, (E_eq_E', (W'r_w, seq_cons_E'))).
       
        clear E_exec_P first_intro_r intro_r_w.
        
        (* It is probably the case that I will need to say that actions E Wr 
           Wr is a Write*)

        (* FROM HERE TRY TO SHOW THE DATA RACE IN E' *)

        destruct E_eq_E' as (eq_act_EE', 
          (eq_ord_EE', (eq_func_EE', (comp_po, (comp_po', (comp_so, comp_so')))))).
                
        destruct (top_sort_order (hb E)(hb_partial_order E)) as (refl_to, trans_to, _);
          fold to in refl_to, trans_to.

        assert(in_r:= lower_eq _ _ (refl_to r)).
        assert(in_w:= lower_eq r _ to_w_r).
        assert(in_Wr:= lower_eq _ _ (trans_to _ _ _ to_Wr_w to_w_r)).

        assert(act'_r:= act_both _ _ (lower_eq_r _) eq_act_EE' _ 
          in_r act_r); clear refl_to trans_to.

        assert(eq_ord_hb_EE':=eq_hb_EE' _ _ _ (proj1 eq_ord_EE')(proj2 eq_ord_EE')
          comp_po comp_po' comp_so comp_so'); clear eq_ord_EE'.

        clear comp_po comp_so comp_so' comp_po'.

        (* From here on we show that the race is W -> w -> r. *)
        assert (race_E':~ (hb E' (W r act_r r_read) w /\ hb E' w r)).
        elim (not_and_or _ _ race_E);
          [intro race_E_Wr_w| intro race_E_w_r].
        apply (or_not_and); left.
        
        red; intro hb'_Wr_w.
        destruct (eq_ord_hb_EE' _ _ in_Wr in_w) as (_, hb_E'_E);
          exact (race_E_Wr_w (hb_E'_E hb'_Wr_w)).

        apply (or_not_and); right.
        
        red; intro hb'_w_r.
        destruct (eq_ord_hb_EE' _ _ in_w in_r) as (hb_E_E', hb_E'_E);
          exact (race_E_w_r (hb_E'_E hb'_w_r)).

        assert(act'_w:= act_both E E' (lower_eq_r r) eq_act_EE' w in_w act_w).
        assert(act'_Wr:= act_both E E' (lower_eq_r r) eq_act_EE' (W r act_r r_read) in_Wr 
          (w_return_actions _ r act_r r_read)).

        (* We will need to say that ~ hb E' w (W r act_r r_read) *)
        assert(not_hb'_w_Wr:~ (hb E' w (W r act_r r_read))).
        (* By absurdity *)
        elim (classic (hb E' w (W r act_r r_read))); [intro hb'_w_Wr|auto].
        destruct (eq_ord_hb_EE' _ _ in_w in_Wr) as (_, hb_E'_E);
          assert (hb_E_w_Wx:=hb_E'_E hb'_w_Wr).
        assert(w_Wr:=sym_eq (
          (top_sort_consistent (hb E)(hb_partial_order E)) w 
          (W r act_r r_read) hb_E_w_Wx to_Wr_w)); contradiction.

        (* We will need to say that ~ hb E' r w *)
        assert(not_hb'_r_w:~ (hb E' r w)).
        
        (* By absurdity *)
        assert(h:= classic (hb E' r w)); elim h; [intro hb'_r_w|auto].
        destruct (eq_ord_hb_EE' _ _ in_r in_w) as (_, hb_E'_E);
          assert (hb_E_r_w:=hb_E'_E hb'_r_w).
        assert(r_w:= (top_sort_consistent (hb E)(hb_partial_order E)) r w hb_E_r_w to_w_r).
        rewrite <- r_w in w_write; rewrite r_read in w_write; discriminate.
        
        clear eq_act_EE' eq_ord_hb_EE'
          eq_func_EE' race_E in_r in_w in_Wr to_Wr_w to_w_r.

        assert(DRF_E':= p_well_synch E' E'_exec_P seq_cons_E').

        (* Almost done, we need to show that there are data races *)
        elim (not_and_or _ _ race_E'); [intro not_hb_Wr_w|intro not_hb_w_r].

        assert(same_var_w_Wr: act_var (W r act_r r_read) = act_var w).
        rewrite <- same_var_r_w; 
          exact (w_on_same_var E r act_r r_read).

        elim(DRF_E' w (W r act_r r_read) act'_w act'_Wr (sym_eq same_var_w_Wr)
          (or_introl (act_kind (W r act_r r_read) = Write) w_write)).
        
        intros; contradiction.
        intros; contradiction.
        
        elim (DRF_E' w r act'_w act'_r (sym_eq same_var_r_w) (or_introl (act_kind r = Write) w_write));
          intuition. 
      Qed.

    End Lemma2_hypothesis.

  End proof.

End ProgramProofs.

Module seq_cons_proof (P:PROGRAM).
  
  Require Import non_empty_list.
  
  Module PP := ProgramProofs(P).
  Import PP.
  Import P.
  
  Section seq_cons.
    
    Variable E:Exec.
    Variable P:Program.
    Hypothesis P_well_synch : correctly_synchronized P.
    Hypothesis E_exec_P : Program_Executions P E.

    Definition hb_strict (E:Exec) :=
      forall (r:Act)(r_act:actions E r)(r_read:act_kind r = Read),
        hb E (W E _ r_act r_read) r.

    Lemma not_seq_not_hb :
      ~ sequential_consistent E -> 
      exists r:Act, 
        forall (r_act:actions E r)(r_read:act_kind r = Read), ~ hb E (W _ _ r_act r_read) r.
    Proof.
      intro not_seq_cons.
      destruct (classic (hb_strict E)) as [hb_str|not_hb_str];
        [contradiction (Lemma2 _ hb_str _ E_exec_P P_well_synch)|].

      destruct (not_all_ex_not _ _ not_hb_str) as (r, not_hb_wr_r); clear not_hb_str.
      exists r; intros r_act r_read hb_wx_x.
      apply not_hb_wr_r; intros r_act' r_read'.
      rewrite <- (proof_irrelevance _ r_read r_read');
        rewrite <- (proof_irrelevance _ r_act r_act'); 
          auto.
    Qed.

    Lemma not_seq_not_hb_ex :
      ~ sequential_consistent E -> 
      exists r:Act, 
        exists r_act:actions E r, 
          exists r_read:act_kind r = Read, ~ hb E (W _ _ r_act r_read) r.
    Proof.
      intro not_seq_cons.
      destruct (classic (hb_strict E)) as [hb_str|not_hb_str];
        [contradiction (Lemma2 _ hb_str _ E_exec_P P_well_synch)|].
      destruct (not_all_ex_not _ _ not_hb_str) as (r, not_hb_wr_r); clear not_hb_str.
      destruct (not_all_ex_not _ _ not_hb_wr_r) as (r_act, rest); clear not_hb_wr_r.
      destruct (not_all_ex_not _ _ rest) as (r_read, not_hb_wr_r); clear rest.
      exists r; exists r_act; exists r_read; auto.
    Qed.
    
    Definition commit_hb_strict (j:Justification.t) : Prop :=
      forall (r:Act)(r_act:actions E r)(r_read:act_kind r = Read), 
        eIn (Justification.comm j) r ->  hb E (W E _ r_act r_read) r.
    
    (*
       * Now I have to prove that if an execution has a read that sees a non-hb
       * write, there should be a Justification where that read appears for the 
       * first time.
       *)
    
    Lemma non_hb_commit : ~ sequential_consistent E ->
      ~ mapProp commit_hb_strict (justification P _ E_exec_P).
    Proof.
      intro not_seq_cons_E.
      destruct (not_seq_not_hb_ex not_seq_cons_E) as (r, (act_r, (r_read, not_hb))).
      destruct (ValidExecutions P E) as (all_act_com, _).
      destruct (nth_Some (justification P E E_exec_P)(Le.le_refl (limit (justification P E E_exec_P)))) 
        as (j, nth_lim_j).
      
      assert (x_just:=nth_limit_In (justification P E E_exec_P) nth_lim_j).
      assert (~ forall (r : Act)(r_act : actions E r)(r_read : act_kind r = Read),
         eIn (Justification.comm j) r -> hb E (W E r r_act r_read) r).
      intro in_comm_hb.
      assert(all_just:= all_act_com E_exec_P j nth_lim_j).
      rewrite all_just in in_comm_hb.
      contradiction (in_comm_hb r act_r r_read act_r).

      exact (singleFail_mapFail j x_just H (l:=justification P E E_exec_P) (P:=commit_hb_strict)).
    Qed.
    
    Lemma hbStrict_seqCons : 
      forall (E:Exec)(E_exec_P: Program_Executions P E), 
        hb_strict E -> sequential_consistent E.
    Proof.
      intros E0 e_exec_p strict.
      apply (Lemma2 E0) with (P:= P); auto.
    Qed.

    Lemma exists_comm_0 : 
      forall (E_exec_P : Ensembles.In (Program_Executions P) E), 
        exists j: Justification.t, nth (justification P E E_exec_P) 0 = Some j.
    Proof.
      intros; destruct (justification P E) as [j|j]; exists j; repeat auto.
    Qed.
    
    Lemma firstCommit_hb : 
      forall (j:Justification.t), 
        nth (justification _ _ E_exec_P) 0 = Some j ->
        hb_strict (Justification.exec j).
    Proof.
      intros j nth_0_j.
      unfold hb_strict; intros r act_r r_read.
      destruct (ValidExecutions P E) as (_, (_, (_, (_, (_, (_, (_, (_, (_, (_, (_, (_ , (_, (r6_0, _)))))))))))))).
      exact (r6_0 E_exec_P j r act_r r_read nth_0_j).
    Qed.
       
    Lemma justification_strict : 
      forall (a : Justification.t) (n:nat),
        nth (justification P E E_exec_P) (S n) = Some a -> 
        exists b : Justification.t, 
          nth (justification P E E_exec_P) n = Some b.
    Proof.
      induction (justification P E E_exec_P); destruct n.
      intros nth_sn_j; simpl in nth_sn_j; discriminate nth_sn_j.
      intros nth_sn_j; simpl in nth_sn_j; discriminate nth_sn_j.
      exists a; simpl; auto.
      simpl;apply IHc; auto.
    Qed.
    
    Definition not_yet_committed (r:Act)(j:Justification.t) : Prop :=
      ~ eIn (Justification.comm j) r.
    
    Section n_seq_E.

      Hypothesis n_seq : ~ sequential_consistent E.
      
      Lemma actions_committed : 
        forall r:Act, actions E r -> ~ mapProp (not_yet_committed r)(justification P E E_exec_P).
      Proof.
        red;clear n_seq.
        intros r act_r not_comm_r.
        destruct (ValidExecutions P E) as (all_act_comm, _).
        
        assert (all_act_comm_E := all_act_comm E_exec_P); clear all_act_comm.
      
        induction (justification _ _ E_exec_P).
        compute in all_act_comm_E.

        assert (all_in_a: Justification.comm a = actions E).
        exact ((all_act_comm_E a)(refl_equal (Some a))); clear all_act_comm_E.
        
        inversion not_comm_r as [j not_com_r_j j_eq_a|].
        destruct not_com_r_j; rewrite all_in_a; auto.
        
        inversion not_comm_r as [|j ls not_com_r j_eq_a]; subst.
        destruct not_com_r as (not_comm_r_a, not_comm_c).
      
        apply IHc; auto.
        
        intros j nth_lim_c_j.
        
        rewrite (S_limit c a) in all_act_comm_E;
          simpl in all_act_comm_E;
            apply all_act_comm_E; auto.
      Qed.
    
    End n_seq_E.
      
    Goal forall (a:Justification.t)(r:Act), ~ not_yet_committed r a -> eIn (Justification.comm a) r.
      intros.
      unfold not_yet_committed in H.
      apply NNPP; auto.
    Qed.
      
    Definition j_E := justification P E E_exec_P.
    
    Inductive lower_com (a b:Justification.t) : Prop :=
    | base : forall n:nat, n < (limit j_E) -> nth j_E n = Some a -> nth j_E (S n) = Some b -> lower_com a b
    | ind : forall c: Justification.t, lower_com a c -> lower_com c b -> lower_com a b.
    
    Lemma lower_included : forall (a b:Justification.t), 
      lower_com a b -> Strict_Included _ (Justification.comm a)(Justification.comm b).
    Proof.
      intros.
      induction H.
      destruct (ValidExecutions P E) as  (_, (_,(_,(_,(com_0, (nest, (_, (_, (_, (_, (_, (_, (r7,_))))))))))))).
      exact (nest _ _ _ n H H0 H1).
      
      unfold Strict_Included in *|-*.
      split.
      
      destruct IHlower_com1 as (IHlower_com1, _).
      destruct IHlower_com2 as (IHlower_com2, _).
      unfold Included in *|-*.
      intros x x_in_a.
      exact (IHlower_com2 _ (IHlower_com1 _ x_in_a)).
      
      red.
      intros.
      rewrite H1 in *|-*.
      assert (Same_set _ (Justification.comm b) (Justification.comm c)).
      destruct IHlower_com1 as (IHlower_com1, _).
      destruct IHlower_com2 as (IHlower_com2, _).
      split; [auto|auto].
      destruct IHlower_com1 as (_, b_not_c).
      exact (b_not_c (Extensionality_Ensembles _ _ _ H2)).
    Qed.
    
    Lemma not_lower_self : forall (j:Justification.t), 
      ~ lower_com j j.
    Proof.
      intros.
      red; intro.
      assert (str_inc_j:=lower_included _ _ H).
      destruct str_inc_j as (_, a); auto.
    Qed.
    
    Lemma lower_antisym : forall (a b:Justification.t), 
      a <> b -> 
      lower_com a b -> 
      ~ lower_com b a.
    Proof.
      red; intros.
      assert (h0:=lower_included _ _ H0).
      assert (h1:=lower_included _ _ H1).
      unfold Strict_Included in * |-.
      destruct h0 as (h01, h02).
      destruct h1 as (h1, _).
      
      assert (same := conj h01 h1).
      fold (Same_set _ (Justification.comm a) (Justification.comm b)) in same.
      
      contradiction (Extensionality_Ensembles _ _ _ same).
    Qed.

    Lemma incl_trans : forall a b c:Ensemble Act, 
      Strict_Included _ a b -> 
      Strict_Included _ b c -> 
      Strict_Included _ a c.
    Proof.
      intros a b c str_inc_a_b str_inc_b_c.
      unfold Strict_Included in *  |- *.
      destruct str_inc_a_b as (inc_a_b, diff_a_b).
      destruct str_inc_b_c as (inc_b_c, diff_b_c).

      split.
      unfold Included in * |- *.
      intros x in_a_x; exact (inc_b_c _ (inc_a_b _ in_a_x)).

      red; intro a_eq_c; subst.
      assert (s_s:=conj inc_b_c inc_a_b).
      contradiction (Extensionality_Ensembles _ _ _ s_s). 
    Qed.
    
    Lemma seq_trans_incl : forall (a b:Justification.t)(n m:nat), 
      n < m ->
      nth j_E n = Some a -> 
      nth j_E m = Some b ->
      Strict_Included _ (Justification.comm a) (Justification.comm b).
    Proof.
      intros a b n m leq_n_m.
      generalize a b; clear a b.
      induction leq_n_m.
      
      intros a b nth_n_a nth_sn_b.
      destruct (ValidExecutions P E) as (_,(_,(_,(_,(_,(nest,_)))))).
      assert(h:=some_Nth _ _ nth_sn_b).
      exact(nest E_exec_P a b n h nth_n_a nth_sn_b).
      
      intros a b nth_n_a nth_sm_b.
      destruct (justification_strict b m nth_sm_b) as (c, nth_m_c).
      assert (h:=IHleq_n_m a c nth_n_a nth_m_c).
      
      assert (Strict_Included _ (Justification.comm c) (Justification.comm b)).
      destruct (ValidExecutions P E) as (_,(_,(_,(_,(_,(nest,_)))))).
      unfold commitsAreNested in nest.
      cut  (m < limit (justification P E E_exec_P)).
      intro lt_m_limit.
      exact(nest E_exec_P c b m lt_m_limit nth_m_c nth_sm_b).
      exact(some_Nth _  _ nth_sm_b).
      exact (incl_trans _ _ _ h H).
    Qed.
    
    Lemma commit_read_write : 
      forall (r:Act)(r_read:act_kind r = Read)(act_r : actions E r)(n:nat)(j:Justification.t),
        nth j_E n = Some j ->
        eIn (Justification.comm j) r ->
        eIn (Justification.comm j) (W E r act_r r_read).
    Proof.
      intros r r_read act_r.
      destruct (ValidExecutions P E) as  (_, (_,(_,(_,(com_0, (_, (_, (_, (_, (_, (_, (_, (r7,_))))))))))))).
      generalize (r7 E_exec_P).
      generalize (com_0 E_exec_P).
      
      induction n.

      intros j nth_0_j.

      rewrite (H _ nth_0_j); compute.
      intro empty_r; destruct empty_r.

      intros j nth_sn_j in_j_r.
      
      destruct(justification_strict j n nth_sn_j) as (e, nth_n_e).
      
      destruct (classic (eIn (Justification.comm e) r)).
      destruct (seq_trans_incl e j n (S n) (Lt.le_lt_n_Sm _ _ (le_n n)) nth_n_e nth_sn_j) as (inc_e_j, _).
      assert (h:=IHn e nth_n_e H1).
      exact (inc_e_j _ h).
      destruct (ValidExecutions P E) as (_,(_,(_,(in_E',_)))).
      assert (in_r_j:=in_E' E_exec_P j r (nth_In _ j_E nth_sn_j) in_j_r).
      
      destruct ((H0 j e n r act_r in_r_j r_read) ((some_Nth j_E) (S n) j nth_sn_j) nth_sn_j nth_n_e).
      
      auto.
      auto.
      
      destruct (seq_trans_incl e j n (S n) (Lt.le_lt_n_Sm _ _ (le_n n)) nth_n_e nth_sn_j) as (inc_e_j, _).
      exact (inc_e_j _ H3).
    Qed.
    
    Lemma commit_read_write_j : 
      forall (r:Act)(r_read:act_kind r = Read)(n:nat)(j:Justification.t)(act_r : actions (Justification.exec j) r),
        nth j_E n = Some j ->
        eIn (Justification.comm j) r ->
        eIn (Justification.comm j) (W (Justification.exec j) r act_r r_read).
    Proof.
      induction n.
      intros.
      destruct (ValidExecutions P E) as  (_, (_,(_,(_,(com_0, _))))).
      rewrite (com_0 _ _ H) in H0.
      destruct H0.
      
      intros.
      destruct(justification_strict j n H) as (e, nth_n_e).
      destruct (classic (eIn (Justification.comm e) r)).
      destruct (ValidExecutions P E) as (_,(_,(in_E,(_,(_,(_,(_,(_,(_,(_,(r5,_))))))))))).
      assert(h:=r5 E_exec_P j e n r (in_E E_exec_P j r (nth_In _ j_E H) H0) act_r r_read ((some_Nth j_E) (S n) j H) H nth_n_e H1).
      
      assert (h1:=(commit_read_write r r_read (in_E E_exec_P j r (nth_In _ j_E H) H0) n e) nth_n_e H1).
      rewrite <- h.
      destruct (seq_trans_incl e j _ _ (Lt.le_lt_n_Sm _ _ (le_n n)) nth_n_e H) as (inc, _).
      exact(inc _ h1).
      
      destruct (ValidExecutions P E) as  (_, (_,(in_E,(_,(_, (_, (_, (_, (_, (_, (_, (_, (r7,_))))))))))))).
      
      assert(h:=r7 E_exec_P j e n r (in_E E_exec_P j r (nth_In _ j_E H) H0) act_r r_read ((some_Nth j_E) (S n) j  H) H nth_n_e H0 H1).
      destruct h.
      destruct (seq_trans_incl e j _ _ (Lt.le_lt_n_Sm _ _ (le_n n)) nth_n_e H) as (inc, _).
      exact (inc _ H2).
    Qed.
    
    Lemma first_race_seq_cons : 
      forall (a : Justification.t) (n:nat),
        nth (justification P E E_exec_P) n = Some a -> 
        ~ commit_hb_strict a ->
        (forall (j : nat) (a : Justification.t),
          j < n ->
          nth (justification P E E_exec_P) j = Some a -> 
          commit_hb_strict a) ->
        hb_strict (Justification.exec a).
    Proof.
      intros a n nth_n_a n_strict_a less_a_C_strict.
      destruct n.
      exact (firstCommit_hb a nth_n_a).
      
      unfold hb_strict.
      intros r act_r_a r_read.
      
      (* Case r \in pred (comes from the less_a_C_strict hip) *)
      destruct (justification_strict a n nth_n_a) as (e, nth_n_e).
      destruct (classic (eIn (Justification.comm e) r)) as [r_in_e|r_not_in_e].
      assert (h:=less_a_C_strict n e (le_n (S n)) nth_n_e).
      destruct (ValidExecutions P E) as (_,(_,(in_E,(in_E',(_,(_,(_,(r2,(_,(_,(r5,(r6, _)))))))))))).
      
      assert (act_r:= (in_E E_exec_P e r (nth_In n j_E nth_n_e) r_in_e)).
      
      assert (h1:= h r act_r r_read r_in_e).
      
      
      assert(h2:=r2 E_exec_P a (W E r act_r r_read) r).
      
      (*       assert(h2:=r2 E_exec_P e (W E r (in_E E_exec_P e r (nth_In n j_E nth_n_e) r_in_e) r_read) r). *)
      fold j_E in h2.
      assert (in_j_a:= nth_In _ _ nth_n_a).
      
      destruct (seq_trans_incl e a n (S n) (Lt.lt_n_Sn n) nth_n_e nth_n_a) as (inc, _).
      assert(com_a_r:=inc _ r_in_e).
      assert (wr_com_a:=commit_read_write r r_read act_r (S n) a nth_n_a com_a_r).
      
      destruct (h2 in_j_a wr_com_a com_a_r) as (h3, h4).
      
      (**)
      
      assert ((W E r act_r r_read) = (W (Justification.exec a) r act_r_a r_read)).
      
      assert (wE_r_e :=commit_read_write r r_read act_r n e nth_n_e r_in_e).
      
      assert (wa_r_e :=commit_read_write_j r r_read (S n) a act_r_a nth_n_a (inc r r_in_e)); 
        clear inc.
      
      exact (r5 E_exec_P a e n r act_r act_r_a r_read (((some_Nth j_E) (S n) a) nth_n_a) nth_n_a nth_n_e r_in_e).
      (**)
      
      rewrite <- H.
      
      apply h3.
      
      intros.
      auto.
      
      (*****************A big case proved************************)
      
      destruct (ValidExecutions P E) as (_,(_,(_,(_,(_,(_,(_,(_,(_,(_,(_,(r6, _)))))))))))).
      
      exact(r6 E_exec_P a e n r act_r_a r_read ((some_Nth j_E) (S n) a nth_n_a) nth_n_a nth_n_e r_not_in_e).
    Qed.


    Lemma data_race : 
      forall (a : Justification.t) (n:nat),
        nth (justification P E E_exec_P) n = Some a -> 
        ~ commit_hb_strict a ->
        hb_strict (Justification.exec a) -> 
        ~ dataRace_free (Justification.exec a).
    Proof.
      intros a n nth_n_a not_com_hb hb_strict_a.
      unfold commit_hb_strict in not_com_hb.
      destruct (not_all_ex_not _ _ not_com_hb) as (r, rest).
      destruct (not_all_ex_not _ _ rest) as (act_r, rest1); clear rest.
      destruct (not_all_ex_not _ _ rest1) as (r_read, rest); clear rest1 not_com_hb.
      assert (r_com_a:=not_imply_elim _ _ rest); 
        assert (not_hb_wr_r:=not_imply_elim2 _ _ rest); 
          clear rest.
      
      destruct (classic (dataRace_free (Justification.exec a))) as [drf_a| H]; [|auto].
      
      (** Prove that (~ hb (Justification.exec a) (W E r act_r r_read) r) per rule 2  **)
      
      assert (not_hba_wr_r: ~ hb (Justification.exec a) (W E r act_r r_read) r).
      destruct (ValidExecutions P E) as (_,(_,(_,(_,(_,(_,(_,(r2,_)))))))).
      
      assert (in_a:= nth_In _ _ nth_n_a).
      assert (wr_com_a:=commit_read_write r r_read act_r n a nth_n_a r_com_a).
      
      destruct (r2 E_exec_P a (W E r act_r r_read) r in_a wr_com_a r_com_a) as (h1, h2).
      
      intro hb_a_wr_r.
      destruct not_hb_wr_r.
      exact(h2 hb_a_wr_r).
      
      (** Prove that (~ hb (Justification.exec a) r (W E r act_r r_read)) 
         per rule 2 and well formedness of E  **)
      
      assert (not_hba_r_w: ~ hb (Justification.exec a) r (W E r act_r r_read)).
      
      assert (~ hb E r (W E r act_r r_read)).
      destruct (ValidExecutions P E) as (_,(_,(_,(_,(_,(_,(_,(r2,(_,(_,(_,(_,(_,(_,(_,wf_E))))))))))))))).
      destruct (wf_E E_exec_P) as (no_inv_E,_); clear wf_E.
      exact (no_inv_E r act_r r_read).
      
      assert (wr_com_a:=commit_read_write r r_read act_r n a nth_n_a r_com_a).
      
      destruct (ValidExecutions P E) as (_,(_,(_,(_,(_,(_,(_,(r2,_)))))))).
      assert (in_j_a := nth_In _ _ nth_n_a).
      destruct (r2 E_exec_P a r (W E r act_r r_read) in_j_a r_com_a wr_com_a) as (E_a,b).
      
      intro hba_r_wr.
      contradiction (b hba_r_wr).
      
      (** Find the race**)
      
      assert (wr_com_a:=commit_read_write r r_read act_r n a nth_n_a r_com_a).
      
      destruct (ValidExecutions P E) as (_, (_, (_, (com_in_E', _)))). 
      assert (in_a:= nth_In _ _ nth_n_a).
      assert(act_a_r:=com_in_E' E_exec_P a r in_a r_com_a).
      assert(act_a_wr:=com_in_E' E_exec_P a (W E r act_r r_read) in_a wr_com_a).
      
      assert(wr_write:=w_returns_writes E r act_r r_read).
      assert(wr_r_same_var:=w_on_same_var E r act_r r_read).
      
      assert (hb_r_wr:=drf_a r (W E r act_r r_read) act_a_r act_a_wr (sym_eq wr_r_same_var) (or_intror (act_kind r = Write) wr_write)).
      
      destruct hb_r_wr; [contradiction|contradiction].
    Qed.
    
    Theorem drf_seq_cons : sequential_consistent E.
      destruct (classic (sequential_consistent E)) as [H|n_seq]; auto.
      
      destruct (firstFail (non_hb_commit n_seq)) as (n, (b,(c,d))).
      destruct (nth_Some j_E b).
      assert (no_hb_str_x:=c _ H).
      
      assert(hb_str_x:=first_race_seq_cons x n H no_hb_str_x d).
      destruct (ValidExecutions P E) as (_, (just_exec, _)).
      assert(x_exec_p:=(just_exec E_exec_P x (nth_In _ _ H))).
      
      assert (seq_cons_x:=hbStrict_seqCons (Justification.exec x) x_exec_p hb_str_x).
      
      assert (data_race_x:=data_race x n H no_hb_str_x hb_str_x).
      
      contradiction (P_well_synch (Justification.exec x) x_exec_p seq_cons_x).
    Qed.
  
  End seq_cons.

End seq_cons_proof.


